import logging
import os
import subprocess
from io import StringIO
from typing import List
from zipfile import ZipFile
import numpy as np
import pandas as pd
import requests
import sqlalchemy

from nexinfosys.ie_imports.data_source_manager import IDataSourceManager, get_dataset_structure
from nexinfosys.models import log_level
from nexinfosys.models.statistical_datasets import DataSource, Database, Dataset, Dimension, CodeList, CodeImmutable

# eng = sqlalchemy.create_engine("monetdb:///demo:", echo=True)
# df = pd.DataFrame(np.random.randn(6, 4), index=pd.date_range('20130101', periods=6), columns=list('ABCD'))
# df.to_sql('test', eng, if_exists='append')

logger = logging.getLogger(__name__)
logger.setLevel(log_level)


"""
To obtain the structure, it is assumed the download process parses the file to obtain dimensions
FAOSTAT needs access to the metadata database

Full database is an iteration?
Also, create full code lists for the different concepts

"""

dataset_csv = """
Code,File,Label
AE,ASTI_Expenditures,ASTI-Expenditures
AF,ASTI_Researchers,ASTI-Researchers
BC,CommodityBalances_Crops,Commodity Balances – Crops Primary Equivalent
BL,CommodityBalances_LivestockFish,Commodity Balances – Livestock and Fish Primary Equivalent
CC,FoodSupply_Crops,Food Supply – Crops Primary Equivalent
CISP,Investment_CountryInvestmentStatisticsProfile_,Country Investment Statistics Profile
CL,FoodSupply_LivestockFish,Food Supply – Livestock and Fish Primary Equivalent
CP,ConsumerPriceIndices,Consumer Price Indices
CS,Investment_CapitalStock,Capital Stock
EA,Development_Assistance_to_Agriculture,Development Flows to Agriculture
EC,Environment_AirClimateChange,Air and Climate change
EE,Environment_Energy,Energy
EF,Environment_Fertilizers,Fertilizers
EI,Environment_Emissions_intensities,Emissions intensities
EK,Environment_LivestockPatterns,Livestock patterns
EL,Environment_LandUse,Land Use
EM,Environment_Emissions_by_Sector,Emissions by Sector
EP,Environment_Pesticides,Pesticides
ES,Environment_Soil,Soil
ET,Environment_Temperature_change,Temperature change
EW,Environment_Water,Water
FA,Food_Aid_Shipments_WFP,Food Aid Shipments (WFP)
FBS,FoodBalanceSheets,Food Balance Sheets
FDI,Investment_ForeignDirectInvestment,Foreign Direct Investment (FDI)
FO,Forestry,Forestry Production and Trade
FS,Food_Security_Data,Suite of Food Security Indicators
FT,Forestry_Trade_Flows,Forestry Trade Flows
GA,Emissions_Agriculture_Crop_Residues,Crop Residues
GB,Emissions_Agriculture_Burning_crop_residues,Burning – Crop Residues
GC,Emissions_Land_Use_Cropland,Cropland
GE,Emissions_Agriculture_Enteric_Fermentation,Enteric Fermentation
GF,Emissions_Land_Use_Forest_Land,Forest Land
GG,Emissions_Land_Use_Grassland,Grassland
GH,Emissions_Agriculture_Burning_Savanna,Burning – Savanna
GI,Emissions_Land_Use_Burning_Biomass,Burning – Biomass
GL,Emissions_Land_Use_Land_Use_Total,Land Use Total
GM,Emissions_Agriculture_Manure_Management,Manure management
GN,Emissions_Agriculture_Energy,Energy Use
GP,Emissions_Agriculture_Manure_left_on_pasture,Manure left on Pasture
GR,Emissions_Agriculture_Rice_Cultivation,Rice Cultivation
GT,Emissions_Agriculture_Agriculture_total,Agriculture Total
GU,Emissions_Agriculture_Manure_applied_to_soils,Manure applied to Soils
GV,Emissions_Agriculture_Cultivated_Organic_Soils,Cultivation of Organic Soils
GY,Emissions_Agriculture_Synthetic_Fertilizers,Synthetic Fertilizers
HS,Indicators_from_Household_Surveys,"Indicators from Household Surveys (gender, area, socieoconomics)"
IC,Investment_CreditAgriculture,Credit to Agriculture
IG,Investment_GovernmentExpenditure,Government Expenditure
LC,Environment_LandCover,Land Cover
MK,Macro-Statistics_Key_Indicators,Macro Indicators
OA,Population,Annual population
OE,Employment_Indicators,Employment Indicators
PA,PricesArchive,Producer Prices – Archive
PD,Deflators,Deflators
PE,Exchange_rate,Exchange rates – Annual
PI,Price_Indices,Producer Price Indices – Annual
PM,Prices_Monthly,Producer Prices – Monthly
PP,Prices,Producer Prices – Annual
QA,Production_Livestock,Live animals
QC,Production_Crops,Crops
QD,Production_CropsProcessed,Crops processed
QI,Production_Indices,Production Indices
QL,Production_LivestockPrimary,Livestock Primary
QP,Production_LivestockProcessed,Livestock Processed
QV,Value_of_Production,Value of Agricultural Production
RA,Inputs_FertilizersArchive,Fertilizers Archive
RF,Inputs_Fertilizers,Fertilizers
RL,Inputs_Land,Land Use
RM,Investment_Machinery,Machinery
RP,Inputs_Pesticides_Use,Pesticides Use
RT,Inputs_Pesticides_Trade,Pesticides Trade
RV,Inputs_FertilizersTradeValues,Fertilizers – Trade Value
RY,Investment_MachineryArchive,Machinery Archive
TA,Trade_LiveAnimals,Live animals
TI,Trade_Indices,Trade indices
TM,Trade_DetailedTradeMatrix,Detailed trade matrix
TP,Trade_Crops_Livestock,Crops and livestock products
"""


class FAOSTAT(IDataSourceManager):  # FAOStat (not AquaStat)
    def __init__(self, datasets_directory, metadata_session_factory, data_engine):
        self._bulk_download_url = "http://fenixservices.fao.org/faostat/static/bulkdownloads/"
        self._datasets = pd.read_csv(StringIO(dataset_csv))
        self._datasets["idx"] = self._datasets["Code"]
        self._datasets.set_index("idx", inplace=True)
        self._datasets_directory = datasets_directory
        if not os.path.isdir(self._datasets_directory):
            os.makedirs(self._datasets_directory, mode=0o770)
        self._metadata_session_factory = metadata_session_factory
        self._data_engine = data_engine

    def get_name(self) -> str:
        """ Source name """
        return self.get_datasource().name

    def get_datasource(self) -> DataSource:
        """ Data source """
        src = DataSource()
        src.name = "FAO"
        src.description = "Food and Agriculture Organization"
        return src

    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        db = Database()
        db.code = "FAOSTAT"
        db.description = "FAOSTAT is the main database provided by FAO. Other databases are not available currently"
        return [db]

    def get_datasets(self, database=None) -> List:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """
        return [(r["Code"], r["Label"], r["Code"]) for i, r in self._datasets.iterrows()]

    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """
        # The manager tries to access the metadata database
        # When this function is called, it is because no metadata is available
        # So, perform the ETL, which includes:
        # * download (if not available),
        # * elaboration of metadata and
        # * import of data
        #
        # TODO Signal a "being updated" status for the dataset
        # TODO Delete all metadata about the dataset.
        ds = get_dataset_structure(self._metadata_session_factory, self, dataset)
        measure_found = False
        for d in ds.dimensions:
            if d.code.lower() == "value":
                measure_found = True
                break
        if not measure_found:
            dd = Dimension()
            dd.code = "Value"
            dd.description = None
            dd.attributes = None
            dd.is_time = False
            dd.is_measure = True
            dd.dataset = ds

        return ds

    def get_dataset_filtered(self, dataset, dataset_params: List[tuple]) -> Dataset:
        """ This method has to consider the last dataset download, to re"""
        # Check for the presence of the dataset in the metadata database
        # Check for the presence of data in the data database (count records)
        ds = get_dataset_structure(self._metadata_session_factory, self, dataset)

        # Read dimensions and their code lists
        dims = {}
        for dim in ds.dimensions:
            dims[dim.code] = dim.code_list.to_dict()
            if dim.is_time:
                interval_start = None
                interval_end = None
                for p in dataset_params.copy():
                    if p.lower() in ["starttime", "startperiod"]:
                        interval_start = dataset_params[p][0]
                        del dataset_params[p]
                    elif p.lower() in ["endtime", "endperiod"]:
                        interval_end = dataset_params[p][0]
                        del dataset_params[p]
                if interval_start and interval_end:
                    dataset_params[dim.code] = [str(i) for i in range(int(interval_start), int(interval_end)+1)]

        # Time

        # Prepare the query (WITH count and columns versions)
        # TODO IMPROVEMENTS: Add GROUP BY if a pivot table is available. Pass the list of dimensions, plus the aggregation as TWO new parameters
        sql = "SELECT * FROM \""+dataset+"\" WHERE "
        first = True
        for col, values in dataset_params.items():
            field_name = col+"_id"
            if first:
                first = False
            else:
                sql += " AND "
            if values:
                if not isinstance(values, list):
                    values = [values]
                sql += "(\""+field_name+"\" IN ("+", ".join(["'"+str(v)+"'" for v in values])+"))"
        # Execute the query store it in pd.DataFrame "df"
        # inspec = sqlalchemy.inspect(self._data_engine)
        # for table_name in inspec.get_table_names():
        #     print(table_name)
        conn = self._data_engine.connect()
        res = conn.execute(sql)
        result_list = res.fetchall()
        conn.close()
        if result_list:
            cols = []
            for f in result_list[0].keys():
                if f.endswith("_id"):
                    # Look for dimension named the same
                    for dim in dims:
                        if dim.lower() == f[:-3].lower():
                            cols.append(dim)
                            break
                else:
                    cols.append(f)
            df = pd.DataFrame(result_list, columns=cols)
            # # Add the descriptions for CODE columns
            # for d, cl in dims.items():
            #     df[d+" (desc.)"] = df[d.lower()+"_id"].map(cl)
        ds.data = df
        return ds

    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass

    def etl_full_database(self, database=None, update=False):
        """ If bulk download is supported, refresh full database """
        pass

    def etl_dataset(self, dataset, update=False) -> str:
        """
        Download a file for an FAO Dataset
        It provokes a total refresh of its Data and Metadata

        :param url:
        :param local_filename:
        :param update:
        :return: String with full file name
        """
        # Find (the exact name is not known before hand)
        file_base = self._datasets.loc[dataset, "File"]
        file_suffix = ["_E_All_Data_(Normalized).zip", "_E_All_Data_(Norm).zip"]
        for i in file_suffix:
            file = self._datasets_directory + file_base + i
            if os.path.isfile(file):
                if update:
                    os.remove(file)
                break
        # TODO Check, if possible, if the file on the Web is newer than the file at the file system
        # Download the file if it is not in the file system
        for i in file_suffix:
            url = self._bulk_download_url + file_base + i
            file = self._datasets_directory + file_base + i
            if not os.path.isfile(file):
                r = requests.get(url, stream=True)
                if r.status_code == 200:
                    # http://stackoverflow.com/questions/15352668/download-and-decompress-gzipped-file-in-memory
                    with open(file, 'wb') as f:
                        for chunk in r.iter_content(chunk_size=1024):
                            if chunk:  # filter out keep-alive new chunks
                                f.write(chunk)
                    break
            else:
                break

        # Read the dataset, with two simultaneous operations: collect metadata (do not write it),
        # and transfer data to the data database
        ds = self._process_single_file(self._datasets_directory, self._datasets.loc[dataset],
                                       file, self._metadata_session_factory, self._data_engine)

        ds.database = self.get_databases()[0]

        return ds

    @staticmethod
    def _process_single_file(datasets_directory, dataset_row, file, engine_metadata, engine_data) -> Dataset:
        """
        Obtain the structure of the dataset
        * Dimensions: each with their code lists
          - Unit as dimension
          - Year is its own code
        * Measures: there is a special column for the measures. Each of the detected codes
        * Measure attributes: add "Flag" and "Note" to each measure
        * Concepts: create also "Concept"s. If they exist, link; if not create them
        * ¿Time Dimension? Not special:
        :param dataset_row: Row of the DataFrame self._datasets with "Code", "File" and "Label" columns 
        :param file: Full file name to process
        :param engine_metadata: Database where Metadata will be added
        :param engine_data: Database where Data will be added
        :return: A Dataset, not persisted
        """
        def get_file_sample(fi):
            # Open the file, get the header
            try:
                zip_file = ZipFile(fi, "r")
                fname = zip_file.namelist()[0]
                tmp = pd.read_csv(zip_file.open(fname), nrows=5, encoding="cp1252")
            except NotImplementedError:
                new_file = datasets_directory + "/" + fname
                if not os.path.exists(new_file):
                    cmd = "/usr/bin/unzip -p \"" + fi + "\" \"" + fname + "\" > \"" + new_file + "\""
                    logging.debug(cmd)
                    subprocess.call(cmd, shell=True)
                fi = new_file
                tmp = pd.read_csv(fi, nrows=5, encoding="cp1252")
            return tmp, fi

        logging.debug("Processing "+dataset_row["Label"]+" dataset ...")

        # pd.DataFrame (first rows of the dataset) + name of file to be read
        tmp, file = get_file_sample(file)
        measures_attributes = ["Flag", "Unit"]

        # Special cases
        if "Year" in tmp.columns and "Year Code" not in tmp.columns:
            add_year_code_column = True
            tmp["Year Code"] = tmp["Year"]
        else:
            add_year_code_column = False

        if "Note" in tmp.columns:
            measures_attributes.append("Note")

        if "Measure" in tmp.columns:
            several_measures = True
        else:
            several_measures = False

        # TODO "ElementGroup" ????

        # Look for pairs of dimensions "<Dim>" and "<Dim> Code", create a dictionary if found, if it does not yet exist
        code_lists = {}  # Dictionary of code lists, each code list's value is a flat dictionary of codes:descriptions
        cols_to_keep = []  # Column names in the DataFrame to KEEP. Used to filter the DataFrame to ONLY these columns
        cols_to_skip = set()  # Columns to be skipped (if there is "Col Code" and "Col", "Col" is skipped)
        col_field_names = []  # Column FIELD names (in the target database)
        dims = set()  # Set of dimensions
        meas = set()  # Set of measures
        attr = set()  # Set of attributes
        # Pairs of columns
        for c in tmp.columns:
            if c.strip().lower().endswith("code"):
                dname = c[:-len("Code")].strip()  # Dimension name without "Code" suffix
                if dname in tmp.columns:
                    cols_to_skip.add(dname)
                    dims.add(dname)
                    cols_to_keep.append(c)
                    col_field_names.append(dname+"_id")
                    if dname.lower() not in code_lists:
                        code_lists[dname.lower()] = {}
        # Columns not in pairs
        for c in tmp.columns:
            if c not in cols_to_skip and c not in cols_to_keep:
                cols_to_keep.append(c)
                col_field_names.append(c)

        # A pass through ALL the file, gather a DICT for each DIMENSION -> CodeList
        count2 = 0
        zip_file = ZipFile(file, "r")
        fname = zip_file.namelist()[0]
        for data in pd.read_csv(zip_file.open(fname), chunksize=100000, encoding="cp1252"):
            logging.debug(f"Chunk: {count2+1}")
            if add_year_code_column:
                data["Year Code"] = data["Year"]
            # For each Dimension, collect DISTINCT values
            for c in data.columns:
                # Look for a pair of dimensions, add codes and descriptions to it
                # For pairs of dimensions, COLLECT DISTINCT pairs (CODE, DESCRIPTION)
                # (this will go to the concept or to the dimension)
                if c.strip().lower().endswith("code"):
                    dname = c[:-len("Code")].strip()
                    if dname in tmp.columns:
                        suffix = c[len(dname):]  # Some times "Code", most of the time " Code"
                        data[dname + suffix] = data[dname+suffix].astype(str)
                        data[dname] = data[dname].astype(str)
                        cl = code_lists[dname.lower()]
                        cl.update(dict(zip(data[dname+suffix].values, data[dname].values)))
            dtypes = {}
            if engine_data:
                logger.debug(f"Writing")
                # Remove unneeded columns
                data = data[cols_to_keep]
                data.reset_index(inplace=True)
                del data["index"]
                # Rename columns
                data.columns = col_field_names
                if not dtypes:
                    for i, t in enumerate(data.dtypes):
                        if t == np.dtype('O'):
                            dtypes[data.columns[i]] = sqlalchemy.types.String(length=16)  # Unicode 16 caracteres
                        elif t == np.dtype('float64'):
                            dtypes[data.columns[i]] = sqlalchemy.types.Float(30)  # Float

                # TODO Possibly, add convenient columns
                # Insert into "Datamart", regenerate table if it exists, only in the first chunk
                data.to_sql(dataset_row["Code"], engine_data, if_exists='append' if count2 > 0 else 'replace', index=False, dtype=dtypes)
                logger.debug("Written")
            count2 += 1

        # ------------------------------
        #     Elaborate the dataset
        # ------------------------------
        ds = Dataset()
        ds.code = dataset_row["Code"]
        ds.description = dataset_row["Label"]
        ds.attributes = {}  # Dataset level attributes? (encode them using a dictionary)
        ds.metadata = None  # Metadata for the dataset SDMX (flow, date of production, etc.)
        ds.database = None  # Assign in the calling function

        # Add Dimensions
        for d in dims:
            if d not in measures_attributes and d != "Value":
                dd = Dimension()
                dd.code = d
                dd.description = d
                dd.attributes = None
                dd.is_time = d in ["Year", "Month"]
                dd.is_measure = False
                dd.code_list = \
                    CodeList.construct(d, d, [""],
                                       [CodeImmutable(k, v, "", []) for k, v in code_lists[d.lower()].items()])
                dd.dataset = ds

        # TODO Add measures
        # All datasets have a single measure, which could be named after the dataset name
        # Measures have attributes: flag, unit and note
        # In one of the datasets there can be several measures
        # The dataset shows dimensions with their codes
        # The Measure column shows
        return ds


# ----------------------------------------------------------------------------------------------------------------------
#   Exploration of dimensions per dataset and codes per dimension
# ----------------------------------------------------------------------------------------------------------------------

def elaborate_dimensions_and_categories_dataframes(dictionary_file, files_directory, engine_metadata, engine_data, output_file_name):
    """
    Elaborate a dataframe containing all datasets in rows, all dimensions in rows, and a check for a cross when a dataset has a dimension
    Elaborate a dataframe per dimension, all dimension values in rows, datasets in columns, showing which codes is using each of the datasets
    Elaborate a PANEL with the result

    :param dictionary_file: A CSV file containing the correspondence between FAOSTAT dataset files and dataset codes and labels
    :param files_directory: A directory containing all FAOSTAT files
    :return: An Excel file with the result
    """
    import glob

    def get_file_name(f0):
        # Find files matching, take the newest one
        fi = None
        file_date = None
        pattern = files_directory + "/" + f0.strip() + "_E_All*.zip"
        for f2 in glob.glob(pattern):
            if not file_date or file_date < os.path.getmtime(f2):
                file_date = os.path.getmtime(f2)
                fi = f2
        if not fi:
            raise Exception(f0 + " NOT FOUND")
        return fi

    def get_file_sample(f0):
        fi = get_file_name(f0)
        # Open the file, get the header
        try:
            fname = ZipFile(fi, "r").namelist()[0]
            tmp = pd.read_csv(fi, nrows=5, encoding="cp1252")
        except NotImplementedError:
            new_file = files_directory + "/" + fname
            if not os.path.exists(new_file):
                cmd = "/usr/bin/unzip -p \""+fi+"\" \""+fname+"\" > \""+new_file+"\""
                logging.debug(cmd)
                subprocess.call(cmd, shell=True)
            fi = new_file
            tmp = pd.read_csv(fi, nrows=5, encoding="cp1252")
        return tmp, fi

    ds = pd.read_excel(dictionary_file)
    m1 = {r["Code"]: r["File"] for i, r in ds.iterrows()}
    m1_rev = {r["File"]: r["Code"] for i, r in ds.iterrows()}
    df1 = pd.DataFrame(index=m1.values())
    d_df = {}  # Dictionary of dataframes, one per dimension
    d_concept = {}  # Dictionary of concepts, each concept's value is a flat dictionary of codes:descriptions
    # TODO A complication would be to prepare for a full code list model, ready to be persisted

    # Just check that all files are in the directory
    f_prefixes = tuple([f for f in ds["File"]])
    _ = [get_file_name(f) for f in f_prefixes]
    # Then check which files are NOT in the list
    pattern = files_directory + "/*_E_All*.zip"
    nf = []

    for f in glob.glob(pattern):
        if not os.path.basename(f).startswith(f_prefixes):
            logging.debug(f + " not registered")
            nf.append(f)

    for f in f_prefixes:
        logging.debug("Processing "+f+" dataset ...")

        # pd.DataFrame (first rows of the dataset) + name of file to be read
        tmp, file = get_file_sample(f)

        # Mark dimensions present in the current Dataset
        # Create an empty dataframe for new dimensions
        #
        lst_cols = []
        lst_col_names = []
        skip_col = set()
        for c in tmp.columns:
            df1.loc[f, c] = "X"
            # Codes for the column, if it is not a special column
            if c not in d_df and c.upper() not in ["VALUE"]:
                d_df[c] = pd.DataFrame(index=m1.values())
            # Look for a pair, create a dictionary in that case, if it does not yet exist
            if c.strip().lower().endswith("code"):
                dname = c[:-len("Code")].strip()
                if dname in tmp.columns:
                    skip_col.add(dname)
                    lst_cols.append(c)
                    lst_col_names.append(dname.lower()+"_id")
                    if dname.lower() not in d_concept:
                        d_concept[dname.lower()] = {}
            elif c.strip() not in skip_col:
                lst_cols.append(c)
                lst_col_names.append(c.lower())

        # A pass through ALL the file, gather a SET for each DIMENSION
        if engine_data:
            pass
            # TODO Remove existing table

        sets_dict = {}
        count2 = 0
        for data in pd.read_csv(file, chunksize=10000, encoding="cp1252"):
            logging.debug("Chunk "+str(count2+1))
            for c in data.columns:
                if c.upper() not in ["VALUE"]:
                    if c not in sets_dict:
                        sets_dict[c] = set()
                    sets_dict[c].update(data[c].astype(str))
                # Look for a pair of dimensions, add codes and descriptions to it
                if c.strip().lower().endswith("code"):
                    dname = c[:-len("Code")].strip()
                    if dname in tmp.columns:
                        suffix = c[len(dname):] # Some times "Code", most of the time " Code"
                        dd = d_concept[dname.lower()]
                        dd.update(dict(zip(data[dname+suffix], data[dname])))
            if engine_data:
                # Remove unneeded columns
                data = data[lst_cols]
                data.reset_index(inplace=True)
                del data["index"]
                # Rename columns
                data.columns = lst_col_names
                # TODO Possibly, add convenient columns
                # Insert into "Datamart", regenerate table if it exists, only in the first chunk
                data.to_sql(m1_rev[f], engine_data, if_exists='append' if count2 > 0 else 'replace', index=False)
            count2 += 1
        # Once file is scanned, write sets for each DIMENSION DataFrame, for the current file
        for c in tmp.columns:
            if c in d_df:
                df = d_df[c]
                for cd in sets_dict[c]:
                    df.loc[f, cd] = "X"

    # Remove empty DataFrames
    d_df = {k: d_df[k] for k in d_df if d_df[k].shape[1] > 0}
    # Write to Excel
    writer = pd.ExcelWriter(output_file_name)
    # Write each worksheet, shorten name to <=31 characters (Excel format limitation)
    df1.to_excel(writer, sheet_name="Dimensions per File")
    _ = [d_df[k].to_excel(writer, sheet_name=k[:31]) for k in d_df]
    if engine_metadata:
        pass
        # TODO Source "FAO", database "FAOSTAT"
        # TODO datasets ds["File"], ds["Code"], ds["Description"]
        # TODO Concepts: create or update all dimensions plus their code lists


"""
 
* Importar conceptos (dimensiones)
  - Con sus listas de códigos
* Importar metadataos de datasets
  - Con sus dimensiones
  - Con sus medidas
  - Apuntando a conceptos
* Importar datos, cada dataset en su tabla. Medidas como dimensión, un campo adicional. Luego, valor y flag. Menos eficiente en espacio, ¿más eficiente en lectura?
* Lectura. Filtrar, obtener dataframe

------------------------------------------------------------------------

* Create MonetDB container "magic-monetdb". Use a Data Volume
* Change monetdb user password
* Python, package pymonetdb and sqlalchemy_monetdb
* Test import
* Install MonetDB JDBC for PyCharm access

"Datamart"

* Import FAO
  - Get structure, write into 
  - Write into relational using SQLAlchemy
  - Dimensions and Measure code as columns
  - Value and Attributes as separate columns
  - If there is a single Measure, ok
  - Map dataset to connection string plus table
* Query FAO
  - Get structure. Dimensions and Measures. Each has 
  - Get dataset
    - Filter
    - Into a Dataframe
    - 
* Make Magic Box use the new approach
  - Refactor
  
External Repositories
* Own Excel
* File system
* Repository
* OECD
* FAOSTAT


Common DS
* Data cube, JSONStat and SDMX style

Storage
* "Datamart"
* 





* All files
  - Delete all data source dimensions
  - 
* A single dataset
  - Delete dataset dimensions
  - Update concept (add codes, update description of codes)
  - Create dimensions
  - Delete dataset table
  - Create dataset table

"""

#a = ZipFile("/home/rnebot/DATOS/FAOSTAT/FoodBalanceSheets_E_All_Data_(Normalized).zip", "r")
#b = a.open(a.namelist()[0])

# EXECUTE THIS

if __name__ == '__main__':
    db_connection_string = "postgresql://postgres:postgres@localhost:5432/FAOSTAT"
    engine_data = sqlalchemy.create_engine(db_connection_string, echo=False)

    elaborate_dimensions_and_categories_dataframes("/home/rnebot/GoogleDrive/AA_MAGIC/FAOSTAT_datasets.xlsx",
                                                   "/home/rnebot/DATOS/FAOSTAT",
                                                   None,
                                                   engine_data,
                                                   "/home/rnebot/GoogleDrive/AA_MAGIC/FAOSTAT_analysis.xlsx")



import json

from nexinfosys.command_generators import Issue, IType
from nexinfosys.command_generators.spreadsheet_command_parsers_v2 import IssueLocation
from nexinfosys.models.musiasem_concepts import Parameter
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects


# TODO Created for V1 commands, it passes directly to V2 because it has been sparsely used
class ParametersCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        issues = []
        sheet_name = self._content["command_name"]
        # Obtain global variables in state
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

        for r, param in enumerate(self._content["items"]):
            name = param["name"]
            p = glb_idx.get(Parameter.partial_key(name))
            if len(p) > 0:
                issues.append(Issue(itype=IType.WARNING,
                                    description="The parameter '" + name + "' has been declared previously. Skipped.",
                                    location=IssueLocation(sheet_name=sheet_name, row=r, column=None)))
                continue
            p = Parameter(name)
            p._default_value = p._current_value = param.get("value")
            p._type = param.get("type")
            p._range = param.get("domain")
            p._description = param.get("description")
            p._group = param.get("group")
            glb_idx.put(p.key(), p)
        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, (dict, list)):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues



import math
from functools import reduce
from operator import add
from typing import Dict, List, Tuple, Optional, Generator
import networkx as nx

from nexinfosys.solving.graph import Node, Weight, EdgeType
from nexinfosys.solving.graph.computation_graph import ComputationGraph
from nexinfosys.command_generators import IType, Issue


class FlowGraph:
    """
    A Flow Graph is a directed graph where:
     - Each node represents a variable to be computed or which value will be given.
     - Each directed edge represents the flowing of data from one node to another.
     - Each edge can have two attributes:
       * Direct weight: a weight affecting the amount of data flowing Top-Down
       * Reverse weight: a weight affecting the amount of data flowing Bottom-Up
    """
    def __init__(self, graph: Optional[nx.DiGraph] = None):
        self._direct_graph = nx.DiGraph()
        self._reverse_graph = nx.DiGraph()

        if graph:
            for u, v, data in graph.edges(data=True):
                self.add_edge(u, v, data["weight"], None)

    def add_edge(self, u: Node, v: Node, weight: Optional[Weight], reverse_weight: Optional[Weight]) -> None:
        """ Add an edge with weight attributes to the flow graph """
        self._direct_graph.add_edge(u, v, weight=weight)
        self._reverse_graph.add_edge(v, u, weight=reverse_weight)

    def edges(self) -> Generator[Tuple[Node, Node, Optional[Weight], Optional[Weight]], None, None]:
        """ Return the edges of the flow graph """
        for u, v, data in self._direct_graph.edges(data=True):  # type: Node, Node, Dict
            yield u, v, data["weight"], self._reverse_graph[v][u]["weight"]

    @property
    def nodes(self):
        """ Return the nodes of the flow graph """
        return self._direct_graph.nodes

    def get_computation_graph(self, structural_graph: Optional[nx.DiGraph] = None) -> Tuple[Optional[ComputationGraph], List[Issue]]:
        """ Get a Computation Graph out of the Flow Graph, after checking the validity (no cycles) and inferring
            missing weights and compute other node attributes.

            :return: the Computation Graph if no errors occurred and a list of messages given by the analysis
        """
        issues = self.analyze_and_complete(structural_graph)

        if IType.ERROR in [e.itype for e in issues]:
            return None, issues

        return self._create_computation_graph(), issues

    def _create_computation_graph(self) -> ComputationGraph:
        """ Create a Computation Graph based on the Flow Graph """
        graph = ComputationGraph()

        for u, v, weight, reverse_weight in self.edges():
            graph.add_edge(u, v, weight, reverse_weight)

        for n, split in self._direct_graph.nodes.data('split'):  # type: Node, bool
            if split:
                graph.mark_node_split(n, EdgeType.DIRECT)

        for n, split in self._reverse_graph.nodes.data('split'):  # type: Node, bool
            if split:
                graph.mark_node_split(n, EdgeType.REVERSE)

        return graph

    def analyze_and_complete(self, structural_graph: Optional[nx.DiGraph] = None) -> List[Issue]:
        """
        It analyzes the flow graph and completes it with inferrable data.

        First, it checks if Flow Graph is DAG (Directed Acyclic Graph), this is, no cycles exist
        Then, it follows these steps:

          How many output edges without weight has the node?
            * More than 1: missing weights (WARNING)
            * Only 1: how many output edges in total?
               * Only 1: is there an opposite weight?
                  * Yes: weight can be inferred as (1.0 / opposite weight) (INFO)
                  * No: weight can be inferred as 1.0 (INFO)
               * More than 1: compute sum of other edges with weight
                  * sum > 1: weight cannot be inferred (WARNING)
                  * sum <= 1: weight can be inferred as (1 - sum) (INFO)

        :return: a list of messages given by the analysis and completion of type INFO, WARNING or ERROR
        """
        issues: List[Issue] = []

        # Checking if graph is acyclic. Just looking at the direct graph is OK.
        # if not nx.algorithms.dag.is_directed_acyclic_graph(self._direct_graph):
        #     print("Cycles detected. The nodes connected in a cycle are: ")
        #     for n in nx.algorithms.cycles.simple_cycles(self._direct_graph):
        #         print(n)
        #     issues.append(Issue(IType.ERROR, 'The graph contains cycles'))
        #     return issues

        for graph, opposite_graph in [(self._direct_graph, self._reverse_graph), (self._reverse_graph, self._direct_graph)]:
            for n in nx.dfs_preorder_nodes(graph):

                graph.nodes[n]['split'] = False

                # Working on output edges only of node 'n'
                all_edges = graph.out_edges(n, data=True)

                if len(all_edges) == 0:
                    continue

                # How many output edges without weight has the node?
                edges_without_weight = [e for e in all_edges if e[2]['weight'] is None]

                if len(edges_without_weight) > 1:
                    str_edges = [f'({e[0]}, {e[1]})' for e in edges_without_weight]
                    issues.append(Issue(IType.WARNING,
                                        f'The following edges don\'t have a weight: {", ".join(str_edges)}'))

                elif len(edges_without_weight) == 1:

                    if len(all_edges) == 1:
                        edge = list(all_edges)[0]
                        u, v, data = edge
                        opposite_weight = opposite_graph[v][u]['weight']
                        if opposite_weight is not None:
                            data['weight'] = Weight(0.0) if opposite_weight == 0.0 else (Weight(1.0) / opposite_weight)
                            issues.append(Issue(IType.INFO,
                                                f'The weight of single output edge "{edge}" could be inferred from '
                                                f'opposite weight "{opposite_weight}"'))
                        elif structural_graph and (structural_graph.has_edge(u, v) or structural_graph.has_edge(v, u)):
                            issues.append(Issue(IType.WARNING,
                                                f'The weight of single output edge "{edge}" will not be inferred '
                                                f'because a structural edge exist'))
                        else:
                            data['weight'] = Weight(1.0)
                            issues.append(Issue(IType.INFO,
                                                f'The weight of single output edge "{edge}" could be inferred '
                                                f'without opposite weight'))
                    else:
                        sum_other_weights = reduce(add, [e[2]['weight'] for e in all_edges if e[2]['weight'] is not None])

                        if sum_other_weights > 1.0:
                            issues.append(Issue(IType.WARNING,
                                                f'The weight of edge "{edges_without_weight[0]}" cannot be inferred, '
                                                f'the sum of other weights is >= 1.0: {sum_other_weights}'))
                        else:
                            edges_without_weight[0][2]['weight'] = Weight(1.0) - sum_other_weights
                            graph.nodes[n]['split'] = True
                            issues.append(Issue(IType.INFO,
                                                f'The weight of edge "{edges_without_weight[0]}" could be inferred '
                                                f'from the sum of other weights'))

                elif len(all_edges) > 1:
                    # All edges have a weight
                    sum_all_weights = reduce(add, [e[2]['weight'] for e in all_edges])
                    if math.isclose(sum_all_weights, 1.0):
                        graph.nodes[n]['split'] = True

        return issues



"""
From memory model, elaborate a spreadsheet in NIS format that generates the same memory model ("idempotent spreadsheet")
"""
import pandas as pd
import toposort
from openpyxl import Workbook
from openpyxl.utils import get_column_letter
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.writer.excel import save_virtual_workbook

from nexinfosys.command_generators.parser_ast_evaluators import ast_evaluator
from nexinfosys.command_generators.parser_field_parsers import string_to_ast, expression_with_parameters
from nexinfosys.common.helper import PartialRetrievalDictionary, create_dictionary, ifnull
from nexinfosys.model_services import State, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import Parameter, List, Hierarchy, HierarchyNode, FactorType, Processor, \
    ProcessorsRelationPartOfObservation, Factor, FactorsRelationScaleObservation, \
    FactorsRelationDirectedFlowObservation, FactorTypesRelationUnidirectionalLinearTransformObservation, \
    ProblemStatement, Taxon


def nis_format_spreadsheet(s: State):
    """

    :param s:
    :return:
    """
    lst = []
    # Obtain principal structures
    glb_idx, p_sets, hierarchies, datasets, mappings = get_case_study_registry_objects(s)
    # (ProcessorScalings, DatasetQry, ListOfCommands, ImportCommands are lost)
    # Metadata
    lst.append(("Metadata", get_metadata(s)))
    # CodeHierarchies
    lst.append(("CodeHierarchies", get_code_hierarchies(glb_idx)))
    # CodeHierarchiesMapping
    # Parameters
    lst.append(("Parameters", get_parameters(glb_idx)))
    # DatasetDef (only imports)
    # DatasetData

    # RefProvenance
    # RefGeographic
    # RefBibliographic
    # InterfaceTypes
    lst.append(("InterfaceTypes", get_interface_types(glb_idx)))
    # ScaleChangeMap
    lst.append(("ScaleChangeMap", get_scale_change_map(glb_idx)))
    # Processors
    lst.append(("BareProcessors", get_processors(glb_idx)))
    # Interfaces
    lst.append(("Interfaces", get_interfaces(glb_idx)))
    # Relationships (scales)
    lst.append(("Relationships Scales", get_scale_relationships(glb_idx)))
    # Relationships (flows)
    lst.append(("Relationships Exchanges", get_exchange_relationships(glb_idx)))
    # ProblemStatement
    lst.append(("ProblemStatement", get_problem_statement(glb_idx)))
    # ScalarBenchmarks
    # ScalarIndicators
    # MatrixIndicators

    # Convert list of pd.DataFrames to Excel workbook
    wb = Workbook(write_only=True)
    for name, df in lst:
        if df.shape[0] < 2:
            continue

        ws = wb.create_sheet(name)
        widths = [0]*(df.shape[1]+1)  # A maximum of 100 columns
        max_columns = 0
        for r in dataframe_to_rows(df, index=False, header=True):
            if len(r) > max_columns:
                max_columns = len(r)
            for i in range(len(r)):
                width = int(len(str(r[i])) * 1.1)
                if width > widths[i]:
                    widths[i] = width

        for i, column_width in enumerate(widths):
            ws.column_dimensions[get_column_letter(i+1)].width = column_width

        for r in dataframe_to_rows(df, index=False, header=True):
            ws.append(r)

    return save_virtual_workbook(wb)


def list_to_dataframe(lst: List) -> pd.DataFrame:
    return pd.DataFrame(data=lst[1:], columns=lst[0])


def get_metadata(s: State) -> pd.DataFrame:
    metadata_dictionary = s.get("_metadata")
    lst = list()
    lst.append(("Case study code", metadata_dictionary.get("case_study_code", "") if metadata_dictionary else ""))
    lst.append(("Case study name", metadata_dictionary.get("case_study_name", "") if metadata_dictionary else ""))
    lst.append(("Title", metadata_dictionary.get("title", "") if metadata_dictionary else ""))
    lst.append(("Subject, topic and/or keywords", metadata_dictionary.get("subject_topic_keywords", "") if metadata_dictionary else ""))
    lst.append(("Description", metadata_dictionary.get("description", "") if metadata_dictionary else ""))
    lst.append(("Geographical level", metadata_dictionary.get("geographical_level", "") if metadata_dictionary else ""))  # A list
    lst.append(("Dimensions", metadata_dictionary.get("dimensions", "") if metadata_dictionary else "")) # A list
    lst.append(("Reference documentation", metadata_dictionary.get("reference_documentation", "") if metadata_dictionary else ""))
    lst.append(("Authors", metadata_dictionary.get("authors", "") if metadata_dictionary else ""))
    lst.append(("Date of elaboration", metadata_dictionary.get("date_of_elaboration", "") if metadata_dictionary else ""))
    lst.append(("Temporal situation", metadata_dictionary.get("temporal_situation", "") if metadata_dictionary else ""))
    lst.append(("Geographical location", metadata_dictionary.get("geographical_situation", "") if metadata_dictionary else ""))
    if metadata_dictionary:
        lst.append(("DOI", metadata_dictionary.get("doi", "")[0] if metadata_dictionary.get("doi") else ""))
    else:
        lst.append(("DOI", ""))
    lst.append(("Language", metadata_dictionary.get("language", "") if metadata_dictionary else ""))
    lst.append(("Restriction level", metadata_dictionary.get("restriction_level", "") if metadata_dictionary else ""))  # A list
    if metadata_dictionary:
        lst.append(("Version", metadata_dictionary.get("version", "")[0] if metadata_dictionary.get("version") else ""))
    else:
        lst.append(("Version", ""))

    # Expand lists
    for i, t in enumerate(lst):
        if isinstance(t[1], list):
            tmp = [t[0]]
            tmp.extend(t[1])
            lst[i] = tuple(tmp)

    return list_to_dataframe(lst)


def get_parameters(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    lst = [["Parameter", "Type", "Domain", "Value", "Group", "Description", "Attributes"]]
    for p in glb_idx.get(Parameter.partial_key()):
        lst.append([p.name, p.type, p._range if p._range else "", ifnull(p.default_value, ""), ifnull(p.group, ""), ifnull(p._description, ""), ""])

    return list_to_dataframe(lst)


def get_code_hierarchies(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    def hierarchy_to_list(nodes: List[HierarchyNode]) -> List[HierarchyNode]:
        nodes_list = []
        if nodes:
            for node in sorted(nodes, key=lambda n: n.name):
                nodes_list.append(node)
                nodes_list.extend(hierarchy_to_list(list(node.get_children())))
        return nodes_list

    lst = [["Source", "HierarchyGroup", "Hierarchy", "Level", "ReferredHierarchy", "Code", "ParentCode", "Label", "Description", "Expression", "GeolocationRef", "GeolocationCode", "Attributes"]]
    for hh in glb_idx.get(Hierarchy.partial_key()):
        if hh.hierarchy_type == Taxon:
            for c in hierarchy_to_list(hh.roots):
                lst.append(["", "", c.hierarchy.name, ifnull(c.level, ""), c.referred_node.hierarchy.name if c.referred_node else "", c.name, c.parent.name if c.parent else "", c.label, c.description, "", "", "", ""])

    return list_to_dataframe(lst)


def get_interface_types(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    lst = [["InterfaceTypeHierarchy", "InterfaceType", "Sphere", "RoegenType", "ParentInterfaceType", "Formula",
            "Description", "Unit", "OppositeSubsystemType", "Attributes"]]
    for itype in glb_idx.get(FactorType.partial_key()):
        hierarchy = itype.hierarchy.name if itype.hierarchy else ""
        roegen_type = itype.roegen_type.name if itype.roegen_type else ""
        parent = itype.parent.name if itype.parent else ""
        lst.append([hierarchy, itype.name, itype.sphere, roegen_type, parent, "",
                    itype.attributes.get("description", ""), itype.attributes.get("unit", ""),
                    itype._opposite_processor_type, ""])
    return list_to_dataframe(lst)


def get_scale_change_map(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    lst = [["OriginHierarchy", "OriginInterfaceType", "DestinationHierarchy", "DestinationInterfaceType", "OriginContext", "DestinationContext", "Scale", "OriginUnit", "DestinationUnit"]]
    for sc in glb_idx.get(FactorTypesRelationUnidirectionalLinearTransformObservation.partial_key()):
        lst.append([sc.origin.hierarchy.name if sc.origin.hierarchy else "", sc.origin.name,
                    sc.destination.hierarchy.name if sc.destination.hierarchy else "", sc.destination.name,
                    sc._origin_context.name if sc._origin_context else "",
                    sc._destination_context.name if sc._destination_context else "",
                    sc._weight, sc._origin_unit, sc._destination_unit])

    return list_to_dataframe(lst)


def get_processors(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    procs = set(glb_idx.get(Processor.partial_key()))  # Unique processors
    d = {}
    for p in procs:
        parent_relations = glb_idx.get(ProcessorsRelationPartOfObservation.partial_key(child=p))
        d[p.ident] = set([p.parent_processor.ident for p in parent_relations])

    lst = [["ProcessorGroup", "Processor", "ParentProcessor", "SubsystemType", "System", "FunctionalOrStructural", "Accounted", "Stock", "Description", "GeolocationRef", "GeolocationCode", "GeolocationLatLong", "Attributes"]]
    # Elaborate a DAG, then iterate over it
    for ident in list(toposort.toposort_flatten(d)):
        p = glb_idx.get(Processor.partial_key(ident=ident))[0]
        gref = p.geolocation.reference if p.geolocation else ""
        gcode = p.geolocation.code if p.geolocation else ""
        pgroup = p.attributes.get("processor_group")
        fname = p.full_hierarchy_names(glb_idx)[0]
        t = [pgroup if pgroup else "", p.name, "", p.attributes.get("subsystem_type", ""), p.attributes.get("processor_system", ""), p.attributes.get("functional_or_structural", ""), p.attributes.get("instance_or_archetype", ""), p.attributes.get("stock"), "", gref, gcode, "", ""]
        parent_relations = glb_idx.get(ProcessorsRelationPartOfObservation.partial_key(child=p))
        if len(parent_relations) == 0:
            lst.append(t)
        else:
            first = True
            for rel in parent_relations:
                t[2] = rel.parent_processor.full_hierarchy_names(glb_idx)[0]
                lst.append(t.copy())
                if first:
                    first = False
                    proper = None
                    for n in p.full_hierarchy_names(glb_idx):
                        if t[2] in n:
                            proper = n
                            break
                    t[1] = proper if proper else t[1]

    return list_to_dataframe(lst)


def get_interfaces(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    # Used to examine "value" as expression, and find variables that are interface names vs parameter names
    params = create_dictionary(data={p.name: None for p in glb_idx.get(Parameter.partial_key())})
    s = State()
    procs = glb_idx.get(Processor.partial_key())
    d = {}
    for p in procs:
        parent_relations = glb_idx.get(ProcessorsRelationPartOfObservation.partial_key(child=p))
        d[p.ident] = set([p.parent_processor.ident for p in parent_relations])

    lst = [["Processor", "InterfaceType", "Interface", "Sphere", "RoegenType", "Orientation", "OppositeSubsystemType", "GeolocationRef", "GeolocationCode", "InterfaceAttributes", "Value", "Unit", "RelativeTo", "Uncertainty", "Assessment", "PedigreeMatrix", "Pedigree", "Time", "Source", "NumberAttributes", "Comments"]]
    # Elaborate a DAG, then iterate over it
    for ident in list(toposort.toposort_flatten(d)):
        p = glb_idx.get(Processor.partial_key(ident=ident))[0]
        ifaces = glb_idx.get((Factor.partial_key(processor=p)))
        iface_names = create_dictionary(data={iface.name: iface for iface in ifaces})
        # Elaborate DAG of Interfaces because of Observations
        d = {}
        for iface in ifaces:
            if iface.ident not in d:
                d[iface.ident] = set()
            for obs in iface.quantitative_observations:
                if obs.relative_factor:
                    d[iface.ident].add(obs.relative_factor.ident)
                # Consider obs.value and non linear dependencies
                if isinstance(obs.value, str):
                    try:
                        value = float(obs.value)
                        unresolved_vars = []
                    except ValueError:
                        ast = string_to_ast(expression_with_parameters, obs.value)
                        evaluation_issues = []
                        value, unresolved_vars = ast_evaluator(exp=ast, state=s, obj=None, issue_lst=evaluation_issues)
                        for unresolved in unresolved_vars:
                            if unresolved not in params:
                                d[iface.ident].add(iface_names[unresolved].ident)

        for ident2 in list(toposort.toposort_flatten(d)):
            iface = glb_idx.get(Factor.partial_key(ident=ident2))[0]
            lst1 = [iface.processor.name, iface.taxon.name, iface.name, iface.sphere, iface.roegen_type.name,
                    iface.orientation, iface.opposite_processor_type, "", "", ""]
            observations = iface.quantitative_observations
            if len(observations) > 0:
                for obs in observations:
                    lst2 = [obs.value, obs.attributes.get("unit", ""), obs.relative_factor.name if obs.relative_factor else "", obs.attributes.get("spread", ""), obs.attributes.get("assessment", ""), obs.attributes.get("pedigree_template", ""), obs.attributes.get("pedigree", ""), obs.attributes.get("time", ""), obs.observer.name if obs.observer else "", "", obs.attributes.get("comments", "")]
                    lst.append(lst1+lst2)
            else:
                lst.append(lst1+["", "", "", "", "", "", "", "", "", ""])

    return list_to_dataframe(lst)


def get_scale_relationships(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    lst = [["OriginProcessors", "OriginInterface", "DestinationProcessors", "DestinationInterface", "BackInterface", "RelationType", "Weight", "ChangeOfTypeScale", "OriginCardinality", "DestinationCardinality", "Attributes"]]
    for rel in glb_idx.get(FactorsRelationScaleObservation.partial_key()):
        lst.append([rel.origin.processor.name, rel.origin.name, rel.destination.processor.name, rel.destination.name, "", "Scale", rel.quantity, "", "", "", ""])

    return list_to_dataframe(lst)


def get_exchange_relationships(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    lst = [["OriginProcessors", "OriginInterface", "DestinationProcessors", "DestinationInterface", "BackInterface", "RelationType", "Weight", "ChangeOfTypeScale", "OriginCardinality", "DestinationCardinality", "Attributes"]]
    for rel in glb_idx.get(FactorsRelationDirectedFlowObservation.partial_key()):
        lst.append([rel.source_factor.processor.name, rel.source_factor.name, rel.target_factor.processor.name, rel.target_factor.name, rel.back_factor.name if rel.back_factor else "", ">", rel.weight, "", "", "", ""])

    return list_to_dataframe(lst)


def get_problem_statement(glb_idx: PartialRetrievalDictionary) -> pd.DataFrame:
    ps = glb_idx.get(ProblemStatement.partial_key())
    if len(ps) == 0:
        ps = [ProblemStatement()]

    lst = [["Scenario", "Parameter", "Value", "Description"]]
    for solving_parameter in ps[0].solving_parameters.items():
        lst.append(["", solving_parameter[0], solving_parameter[1], ""])

    for scenario, params in ps[0].scenarios.items():
        for k, v in params.items():
            lst.append([scenario, k, v, ""])

    return list_to_dataframe(lst)





"""
High level functions for different use cases:

* Start a case study from scratch
* Continue working on a case study. Implies saving the case study. A status flag for the case study (in elaboration, ready, publishable, ...)
* Play with an existing case study. CRUD elements, modify parameters, solve, analyze (read), ....
* Start an anonymous case study from scratch. It will not be saved
* Create case study from an existing one, as new version ("branch") or totally new case study
  *
* Export case study for publication
* Export objects
* Analyze case study
* Import case study
* Import objects

"""

# All functions use the interactive session
# They need to open a session to which commands can be added


def close_use_case():
    """ Many of the use cases open a"""
    return None


def start_case_study():
    """
    No parameters
    :return:
    """
    return None


def continue_working_case_study():
    """
    Need the code of the case study. Or the version.
    Closed version cannot be opened this way
    :return:
    """


def play_with_existing_case_study():
    """
    Open case study
    The session will not be persisted
    Or would it be just a normal session without final "persist"?
    :return:
    """


def start_anonymous_case_study():
    """
    Open case study
    Without identity
    Do not allow saving, only during session
    ¿What's the difference with start case study and not save it?
    Maybe when opening a case study WARN that, if the user is not identified, nothing can be persisted
    Also, think about limiting command_executors regarding resources consumption (memory AND CPU)

    :return:
    """


"""
Given a State, elaborate an equivalent RDF or OWL file which can be exploited outside

To view it (suggestions):

* https://github.com/essepuntato/LODE
* http://vowl.visualdataweb.org/webvowl.html


RDF Basics:
Triple: (s,p,o)
s (subject), ->p (predicate or property)-> o (object). Node-arc-Node
s: URIRef, Blank Node
p: URIRef
o: URIRef, Blank Node, Literal

Where...
* URIRef. A full URL or a QName (a schema prefix, ":", a local name)
* Blank nodes. Nodes with no intrinsic name. Blank Node Identifier (an ad-hoc identifier)
* URI references URIRef. Vocabulary is "URI based"
* Literals. Literal (unicode) + datatype or language tag (RFC 3066)

Concept: "Nodes": set of Subjects and Objects of a graph

Concept: "Ground RDF": graph with no Blank nodes

Concept: "Name": URIRef or Literal

RDF Principles:
* XML-based syntax
* XML-schema datatypes

Resource. A COMPACT way to access a SUBJECT to read or write its properties
Domain
Range

NamespaceManager: associates Namespace with Prefixed
Each RDFLIB Graph has a "namespace_manager":
 - Populated when reading an RDF
 - New can be added using graph.bind()
 - SPARQL queries can be initialized using "initNs" parameter in method "query" of a GRAPH

Persistence:
 - Memory
 - Sleepycat (BSDDB or BSDDB3)
 - SPARQLStore
 - SPARQLUpdateStore

"""
import os
import sys
from enum import Enum
import owlready2
import io

from rdflib.resource import Resource

from nexinfosys.model_services import State, get_case_study_registry_objects

from nexinfosys.model_services.workspace import execute_file_return_issues
from nexinfosys.initialization import prepare_and_reset_database_for_tests
from nexinfosys.models.musiasem_concepts import FactorType, Processor, Factor, FactorsRelationScaleObservation, \
    FactorsRelationDirectedFlowObservation, ProcessorsRelationPartOfObservation, \
    FactorTypesRelationUnidirectionalLinearTransformObservation, FactorQuantitativeObservation
from nexinfosys.serialization import deserialize_state, serialize_state


class Ontology(Enum):
    # Basic ontologies
    XML = 'http://www.w3.org/XML/1998/namespace'
    XSD = 'http://www.w3.org/2001/XMLSchema#'
    RDF = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
    RDFS = 'http://www.w3.org/2000/01/rdf-schema#'
    OWL = 'http://www.w3.org/2002/07/owl#'
    PROV = 'http://www.w3.org/ns/prov#'
    DC = 'http://purl.org/dc/elements/1.1/'
    DCT = 'http://purl.org/dc/terms/'
    SCHEMA = 'http://schema.org/'

    # special interest namespaces
    SKOS = 'http://www.w3.org/2004/02/skos/core#'
    FOAF = 'http://xmlns.com/foaf/0.1/'
    MA = 'http://www.w3.org/ns/ma-ont#'
    SIOC = 'http://rdfs.org/sioc/ns#'
    PO = 'http://purl.org/ontology/po/'

    # lexical namespaces
    LEMON = 'http://lemon-model.net/lemon#'
    LEXINFO = 'http://www.lexinfo.net/ontology/2.0/lexinfo#'
    LANG = 'http://id.loc.gov/vocabulary/iso639-1/'

    # wikidata, dbpedia, geo
    DBO = 'http://dbpedia.org/ontology/'
    GEO = 'http://www.opengis.net/ont/geosparql#'
    WDT = 'http://www.wikidata.org/prop/direct/'
    WD = 'http://www.wikidata.org/entity/'
    P = 'http://www.wikidata.org/prop/'
    PS = 'http://www.wikidata.org/prop/statement/'
    PQ = 'http://www.wikidata.org/prop/qualifier/'
    GN = 'http://sws.geonames.org/'

    @classmethod
    def to_fully_qualified(cls, prefix: str) -> str:
        """ Look up short prefix and return fully-qualified URL if known.
        :param prefix: the prefix to be resolved.
        :returns the fully-qualified URL or the prefix if unknown.
        """
        prefix = prefix.upper()
        if hasattr(Ontology, prefix):
            return getattr(cls, prefix).value
        return prefix

    @classmethod
    def to_prefix(cls, uri: str) -> str:
        """
        """
        try:
            return Ontology(uri).name.lower()
        except Exception as e:
            pass


NAMESPACES = {item.value: item.name.lower() for item in Ontology}

PREFIXES = '\n'.join([''] + ['PREFIX {value}: <{key}>'.format(value=item.name.lower(),
                                                              key=item.value)
                             for item in Ontology])


def to_fully_qualified(attribute: str) -> str:
    """ QName originates from the XML world, where it is used to reduce I/O by shortening
    namespaces (e.g. http://www.weblyzard.com/wl/2013#) to a prefix (e.g. wl)
    followed by the local part (e.g. jonas_type). The namespace-prefix relations
    are thereby defined in the XML-Head. This relation mapping does not exist in JSON,
    which is why we have to have the full qualified name (namespace + local part)
    to define an attribute. While the more readable way would be to just use the URI,
    the official standardized format is {namespace}localpart, having the major
    advantage of non-ambiguous namespace identification. Further, Java expects Qnames
    in this annotation format, which enables to simply use Qname.valueOf(key).
    :param attribute: the attribute to resolve
    :returns a fully-qualified version of the input attribute.
    """
    if len(attribute.split(':')) <= 1 or attribute.startswith('{'):
        return attribute

    namespace, attr_name = attribute.split(':')
    return '{%s}%s' % (Ontology.to_fully_qualified(namespace), attr_name)


def prefix_uri(uri: str, allow_partial: bool=False) -> str:
    """ Replace a sub-path from the uri with the most specific prefix as defined
    in the Namespace.
    :param uri: The URI to modify.
    :type uri: str
    :returns: The modified URI if applicable
    :rtype: str
    """
    if not uri.startswith('http'):
        return uri
    # replace most specific/longest prefix, hence sorted
    for namespace in sorted(list([ns.value for ns in Ontology]), key=len, reverse=True):
        if namespace in uri:
            replaced = uri.replace(
                namespace, '{}:'.format(Ontology.to_prefix(namespace)))
            if '/' in replaced or '#' in replaced:
                if not allow_partial:
                    # slashes or hashes in prefixed URIs not allowed
                    continue
            return replaced
    return uri


def replace_prefix(uri):
    """ Replace a prefix with the fully-qualified namespace URL.
    :param uri: The URI to modify.
    :type uri: str
    :returns: The modified URI if applicable
    :rtype: str
    """
    for namespace in sorted(list(NAMESPACES.keys()), key=len, reverse=True):
        prefix = '{}:'.format(NAMESPACES[namespace])
        if uri.startswith(prefix):
            return uri.replace(prefix, namespace)
    return uri


def parse_language_tagged_string(value: str) -> tuple:
    """ Check if a string value has a language tag @xx or @xxx
    and returns the string without the value tag and language
    as tuple. If no language tag -> language is None
    :param value
    :returns:
    """
    lang = None
    if len(value) > 1 and value[0] == value[-1] == '"':
        value = value[1:-1]
    if len(value) > 6 and value[-6] == '@':
        lang = value[-5:]
        value = value[:-6]
    elif len(value) > 3 and value[-3] == '@':
        lang = value[-2:]
        value = value[:-3]
    elif len(value) > 4 and value[-4] == '@':
        lang = value[-3:]
        value = value[:-4]
    return value, lang


def generate_rdf_from_object_model(state: State):
    """
    Using the base ontology "musiasem.owl", generate a file

    :param state:
    :return:
    """
    onto = owlready2.get_ontology("file:////home/rnebot/Dropbox/nis-backend/nexinfosys/ie_exports/musiasem.owl").load()
    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
    # onto = get_ontology("file:///musiasem.owl").load()

    m = {}  # Map from objects of the model in memory to the equivalent OWL individuals

    # InterfaceTypes
    fts = glb_idx.get(FactorType.partial_key())
    for ft in fts:
        ft_individual = onto.InterfaceType(ft.name)
        ft_individual.label = ft.name
        ft_individual.has_roegen_type = ft.roegen_type
        ft_individual.has_sphere = ft.sphere
        ft_individual.has_processor_type = ft.opposite_processor_type
        ft_individual.has_unit = ft.unit
        if ft.range:
            tmp = onto.InterfaceRange()
            tmp.has_interval = ft.range
            tmp.has_interval_unit = ft.range_unit
            ft_individual.has_range = tmp

        ft_individual.has_interface = []

        m[ft] = ft_individual

    # Processors and Systems
    procs = glb_idx.get(Processor.partial_key())
    for proc in procs:
        proc_individual = onto.Processor(proc.full_hierarchy_names(glb_idx)[0])
        proc_individual.has_interface = []
        proc_individual.has_part_directly = []
        proc_individual.has_system = proc.processor_system  # Valor string
        proc_individual.has_subsystem_type = proc.subsystem_type  # Valor entre una lista (enum)
        proc_individual.has_processor_type = proc.functional_or_structural  # Valor entre una lista (enum)
        proc_individual.is_accounted = proc.instance_or_archetype  # Boolean
        proc_individual.has_level = proc.level  # Valor string
        if proc.geolocation:
            tmp = onto.Geolocation()
            proc.geolocation.code
            proc.geolocation.ref
            proc_individual.location = tmp
        m[proc] = proc_individual

    # Interfaces and Quantities
    fs = glb_idx.get(Factor.partial_key())
    for f in fs:
        f_individual = onto.Interface()
        f_individual.label = f.name
        f_individual.has_roegen_type = [f.roegen_type]
        f_individual.has_sphere = [f.sphere]
        f_individual.has_orientation = [f.orientation]
        f_individual.has_processor_type = [f.opposite_processor_type]
        f_individual.has_unit = [f.unit]
        if f.range:
            tmp = onto.InterfaceRange()
            tmp.has_interval = [f.range]
            tmp.has_interval_unit = [f.range_unit]
            f_individual.has_range = tmp

        f_individual.has_qqobservation = []
        m[f] = f_individual

        # Link Interface -> AND <- Processor
        f_individual.has_processor = [m[f.processor]]
        m[f.processor].has_interface.append(f_individual)

        # Link Interface -> AND <- InterfaceType
        f_individual.has_interface_type = [m[f.taxon]]
        m[f.taxon].has_interface.append(f_individual)

        # Qualified Quantities
        for q in f.observations:
            if isinstance(q, FactorQuantitativeObservation):
                q_individual = onto.QualifiedQuantity()
                q_individual.has_value = [q.value]
                q_individual.has_unit = [q.unit]
                q_individual.has_time = [q.time]
                q_individual.has_source = [q.source]

                m[q] = q_individual
                # Link Interface -> AND <- QualifiedQuantity
                q_individual.has_interface = [m[f]]
                m[f].has_qqobservation.append(q_individual)

    # Relationships

    #  Part-Of
    part_ofs = glb_idx.get(ProcessorsRelationPartOfObservation.partial_key())
    for po in part_ofs:
        po_individual = onto.PartOf()
        po_individual.has_parent = [m[po.parent_processor]]
        po_individual.has_child = [m[po.child_processor]]
        po_individual.has_weight = 1

        m[po.child_processor].has_parent.append(m[po.parent_processor])
        m[po.parent_processor].has_part_directly.append(m[po.child_processor])

    #  Exchange
    exchanges = glb_idx.get(FactorsRelationDirectedFlowObservation.partial_key())
    for ex in exchanges:
        ex_individual = onto.SequentialFlowLinkage()
        ex_individual.has_origin = [m[ex.source]]
        ex_individual.has_child = [m[ex.target]]
        ex_individual.has_weight = ex.weight
        # ex_individual.has_observer

    #  Scale
    scale = glb_idx.get(FactorsRelationScaleObservation.partial_key())
    #  ScaleChange
    scale_change = glb_idx.get(FactorTypesRelationUnidirectionalLinearTransformObservation.partial_key())

    # Processor Properties
    #  System (maybe a Class)
    #  Geolocation (maybe a Class)
    #  is_functional (boolean)
    #  is_structural (boolean)
    #  is_notional (boolean)
    #  ...
    # InterfaceType Properties
    #  RoegenType (NOT a class)
    #
    # Interface Properties
    #  Orientation
    #  Quantification (a Class)
    #
    # Relationship Properties
    #  Source (another individual)
    #  Target (another individual)
    #  Weight (property)

    # Formal 2 Semantic
    #  Datasets
    #  Hierarchies
    #  Mappings

    out = io.BytesIO()
    onto.save(out, format="n3")

    s = out.getvalue().decode("utf-8")
    with open("/home/rnebot/Downloads/prueba.txt", "wt") as f:
        f.write(s)

    return s


def ensure_dir(file_path):
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory):
        os.makedirs(directory)


if __name__ == '__main__':
    # graph = Graph(identifier="g1")
    # graph.parse("/home/rnebot/Dropbox/nis-backend/nexinfosys/ie_exports/musiasem.owl", format="xml")
    # # graph.serialize("", format="n3")
    # n = Namespace("http://ngd.org/")
    # person = Resource(graph, n["person#p1"])
    # bn = BNode()
    # aref = URIRef("Hola")
    # # graph.add((aref, n.Code, Literal("x01")))
    # result = graph.parse("http://www.w3.org/2000/10/swap/test/meet/blue.rdf")
    # for p in graph:
    #     print(f"{type(p)}{p}")
    #
    # # graph.add()  # graph.input(file_input1) or graph.load(...) or graph.parse()
    # graph.close()
    # graph.add((aref, n.Code, Literal("x01")))
    # for p in graph:
    #     print(f"{type(p)}{p}")
    # print(f"{Literal('01', datatype=XSD.integer)!=Literal('1', datatype=XSD.integer)}")
    # print(f"{n.Person},{n['hola%20adios']}")
    # print("Hola")
    # sys.exit(1)

    input_file = "/home/rnebot/Dropbox/nis-internal-tests/pruebaSoyBeansNIS.xlsx"
    state_file = "/home/rnebot/Dropbox/nis-internal-tests/pruebaSoyBeansNIS.serialized_state"
    if not os.path.exists(state_file):
        prepare_and_reset_database_for_tests(prepare=True)
        isess, issues = execute_file_return_issues(input_file, generator_type="spreadsheet")
        ensure_dir(state_file)
        # Save state
        s = serialize_state(isess.state)
        with open(state_file, "wb") as f:
            f.write(s)
        state = isess.state
    else:
        with open(state_file, "rb") as f:
            s = f.read()
            state = deserialize_state(s)

    generate_rdf_from_object_model(state)




from typing import Dict, Any, Optional

from nexinfosys.command_executors import BasicCommand, CommandExecutionError, subrow_issue_message
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.command_generators import IType
from nexinfosys.command_generators.parser_ast_evaluators import dictionary_from_key_value_list
from nexinfosys.common.helper import strcmp
from nexinfosys.ie_exports.geolayer import read_geojson
from nexinfosys.models.musiasem_concepts import ProcessorsSet, ProcessorsRelationPartOfObservation, \
    Processor, \
    Geolocation, GeographicReference
from nexinfosys.models.musiasem_concepts_helper import find_or_create_processor, \
    obtain_name_parts, find_processors_matching_name


def get_object_view(d):
    class objectview(object):
        def __init__(self, d2):
            self.__dict__ = d2
    return objectview(d)


class ProcessorsCommand(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))

    # def _execute(self, state: "State"):
    #     """
    #     Create empty Processors, potentially related with PartOf relationship
    #
    #     The Name of a new Processor can be:
    #     * Simple name
    #     * Hierarchical name
    #     * Clone previously declared processor (and children), where "previously declared processor" can be simple or complex
    #
    #     If name is hierarchic, a formerly declared processor is assumed.
    #     If a parent processor is specified, the processor will be related to the new parent (if parents are different -> multiple functionality)
    #     In a hierarchy, children inherit interfaces from parents
    #                     if parents do not have interfaces, do they "receive" interfaces from immediate child?
    #                     - Only if children have them already (separate copy)
    #                     - If children do not have them already (connected copy)
    #
    #     If CLONE(<existing processor>) is specified, a copy of the <existing processor> (and children) will be created.
    #     In this case, the Parent field is mandatory (if not, no unique name will be available). For a pair "parent processor - clone processor", this operation is unique (unless, a naming syntax for the copy is invented)
    #     If child processors had interface, copy them into the parent processor (the default move is downwards, from parent to child)
    #
    #     Later, when interfaces are attached to processors,
    #     :param state:
    #     :return:
    #     """
    #     def parse_and_unfold_line(item):
    #         # Consider multiplicity because of:
    #         # - A dataset (only one). First a list of dataset concepts used in the line is obtained.
    #         #   Then the unique tuples formed by them are obtained.
    #         # - Processor name.
    #         #   - A set of processors (wildcard or filter by attributes)
    #         #   - A set of interfaces (according to another filter?)
    #         # - Multiple types of relation
    #         # - Both (first each dataset record applied -expanded-, then the name evaluation is applied)
    #         # - UNRESOLVED: expressions are resolved partially. Parts where parameters
    #         # expressions depending on parameters. Only the part of the expression depending on varying things
    #         # - The processor name could be a concatenation of multiple literals
    #         #
    #         # Look for multiple items in r_source_processor_name, r_source_interface_name,
    #         #                            r_target_processor_name, r_target_interface_name
    #         if item["_complex"]:
    #             asts = parse_line(item, fields)
    #             if item["_expandable"]:
    #                 # It is an expandable line
    #                 # Look for fields which are specified to be variable in order to originate the expansion
    #                 res = classify_variables(asts, datasets, hh, parameters)
    #                 ds_list = res["datasets"]
    #                 ds_concepts = res["ds_concepts"]
    #                 h_list = res["hierarchies"]
    #                 if len(ds_list) >= 1 and len(h_list) >= 1:
    #                     issues.append(create_issue(IType.ERROR, "Dataset(s): "+", ".join([d.name for d in ds_list])+", and hierarchy(ies): "+", ".join([h.name for h in h_list])+", have been specified. Either a single dataset or a single hiearchy is supported."))
    #                     return
    #                 elif len(ds_list) > 1:
    #                     issues.append(create_issue(IType.ERROR, "More than one dataset has been specified: "+", ".join([d.name for d in ds_list])+", just one dataset is supported."))
    #                     return
    #                 elif len(h_list) > 1:
    #                     issues.append(create_issue(IType.ERROR, "More than one hierarchy has been specified: " + ", ".join([h.name for h in h_list])+", just one hierarchy is supported."))
    #                     return
    #                 const_dict = obtain_dictionary_with_literal_fields(item, asts)
    #                 if len(ds_list) == 1:
    #                     # If a measure is requested and not all dimensions are used, aggregate or
    #                     # issue an error (because it is not possible to reduce without aggregation).
    #                     # If only dimensions are used, then obtain all the unique tuples
    #                     ds = ds_list[0]
    #                     measure_requested = False
    #                     all_dimensions = set([c.code for c in ds.dimensions if not c.is_measure])
    #                     for con in ds_concepts:
    #                         for c in ds.dimensions:
    #                             if strcmp(c.code, con):
    #                                 if c.is_measure:
    #                                     measure_requested = True
    #                                 else:  # Dimension
    #                                     all_dimensions.remove(c.code)
    #                     only_dimensions_requested = len(all_dimensions) == 0
    #
    #                     if measure_requested and not only_dimensions_requested:
    #                         issues.append(create_issue(IType.ERROR, "It is not possible to use a measure if not all dimensions are used (cannot assume implicit aggregation)"))
    #                         return
    #                     elif not measure_requested and not only_dimensions_requested:
    #                         # TODO Reduce the dataset to the unique tuples (consider the current case -sensitive or not-sensitive-)
    #                         data = None
    #                     else:  # Take the dataset as-is
    #                         data = ds.data
    #
    #                     for row in data.iterrows():
    #                         item2 = const_dict.copy()
    #
    #                         d = {}
    #                         for c in ds_concepts:
    #                             d["{" + ds.code + "." + c + "}"] = row[c]
    #                         # Expand in all fields
    #                         for f in fields:
    #                             if f not in const_dict:
    #                                 # Replace all
    #                                 string = item[f]
    #                                 # TODO Could iterate through the variables in the field (not IN ALL FIELDS of the row)
    #                                 for item in sorted(d.keys(), key=len, reverse=True):
    #                                     string = re.sub(item, d[item], string)
    #                                 item2[f] = string
    #                         # Now, look for wildcards where it is allowed
    #                         r_source_processor_name = string_to_ast(processor_names, item2.get("source_processor", None))
    #                         r_target_processor_name = string_to_ast(processor_names, item2.get("target_processor", None))
    #                         if ".." in r_source_processor_name or ".." in r_target_processor_name:
    #                             if ".." in r_source_processor_name:
    #                                 source_processor_names = obtain_matching_processors(r_source_processor_name, all_processors)
    #                             else:
    #                                 source_processor_names = [r_source_processor_name]
    #                             if ".." in r_target_processor_name:
    #                                 target_processor_names = obtain_matching_processors(r_target_processor_name, all_processors)
    #                             else:
    #                                 target_processor_names = [r_target_processor_name]
    #                             for s in source_processor_names:
    #                                 for t in target_processor_names:
    #                                     item3 = item2.copy()
    #                                     item3["source_processor"] = s
    #                                     item3["target_processor"] = t
    #                                     print("Multiple by dataset and wildcard: " + str(item3))
    #                                     yield item3
    #                         else:
    #                             print("Multiple by dataset: " + str(item3))
    #                             yield item2
    #                 elif len(h_list) == 1:
    #                     pass
    #                 else:  # No dataset, no hierarchy of categories, but still complex, because of wildcards
    #                     wildcard_in_source = ".." in item.get("source_processor", "")
    #                     wildcard_in_target = ".." in item.get("target_processor", "")
    #                     if wildcard_in_source or wildcard_in_target:
    #                         r_source_processor_name = string_to_ast(processor_names, item.get("source_processor", None))
    #                         r_target_processor_name = string_to_ast(processor_names, item.get("target_processor", None))
    #                         if wildcard_in_source:
    #                             source_processor_names = obtain_matching_processors(r_source_processor_name, all_processors)
    #                         else:
    #                             source_processor_names = [item["source_processor"]]
    #                         if wildcard_in_target:
    #                             target_processor_names = obtain_matching_processors(r_target_processor_name, all_processors)
    #                         else:
    #                             target_processor_names = [item["target_processor"]]
    #                         for s in source_processor_names:
    #                             for t in target_processor_names:
    #                                 item3 = const_dict.copy()
    #                                 item3["source_processor"] = s
    #                                 item3["target_processor"] = t
    #                                 print("Multiple by wildcard: "+str(item3))
    #                                 yield item3
    #                     else:
    #                         # yield item
    #                         raise Exception("If 'complex' is signaled, it should not pass by this line")
    #         else:
    #             # print("Single: "+str(item))
    #             yield item
    #
    #     issues = []
    #     glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
    #     command_name = self._content["command_name"]
    #
    #     # CommandField definitions for the fields of Interface command
    #     fields: Dict[str, CommandField] = {f.name: f for f in get_command_fields_from_class(self.__class__)}
    #     # Obtain the names of all parameters
    #     parameters = [p.name for p in glb_idx.get(Parameter.partial_key())]
    #     # Obtain the names of all processors
    #     all_processors = get_processor_names_to_processors_dictionary(glb_idx)
    #
    #     # Process parsed information
    #     for line in self._content["items"]:
    #         row = line["_row"]
    #         for sub_line in parse_and_unfold_line(line):
    #             self._process_row(sub_line)
    #
    #     return issues, None  # Issues, Output

    def _process_row(self, field_values: Dict[str, Any], subrow=None) -> None:

        # Transform text of "attributes" into a dictionary
        if field_values.get("attributes"):
            try:
                field_values["attributes"] = dictionary_from_key_value_list(field_values["attributes"], self._glb_idx)
            except Exception as e:
                self._add_issue(IType.ERROR, str(e)+subrow_issue_message(subrow))
                return
        else:
            field_values["attributes"] = {}

        # Process specific fields

        # Obtain the parent: it must exist. It could be created dynamically but it's important to specify attributes
        if field_values.get("parent_processor"):
            try:
                parent_processor = self._get_processor_from_field("parent_processor")
                # parents = find_processors_matching_name(parent_processor)
                # if len(parents) > 1:
                #     self._add_issue(IType.WARNING,
                #                     f"Parent processor '{parent_processor}' not unique. Matches: {', '.join(p.hierarchical_names[0] for p in parents)}. Skipped." + subrow_issue_message(subrow))
                #     return
            except CommandExecutionError:
                self._add_issue(IType.ERROR, f"Specified parent processor, '{field_values.get('parent_processor')}', does not exist"+subrow_issue_message(subrow))
                return
        else:
            parent_processor = None

        behave_as_processor: Optional[Processor] = None
        if field_values.get("behave_as_processor"):
            try:
                behave_as_processor = self._get_processor_from_field("behave_as_processor")
            except CommandExecutionError:
                self._add_issue(IType.WARNING, f"Specified 'behave as' processor, '{field_values.get('behave_as_processor')}', does not exist, value ignored"+subrow_issue_message(subrow))

        # Find or create processor and REGISTER it in "glb_idx"
        # TODO Now, only Simple name allowed
        # TODO Improve allowing hierarchical names, and hierarchical names with wildcards
        pgroup = field_values.get("processor_group")

        # Get internal and user-defined attributes in one dictionary
        attributes = {c.name: field_values[c.name] for c in self._command_fields if c.attribute_of == Processor}
        attributes.update(field_values["attributes"])
        attributes["processor_group"] = pgroup

        # Needed to support the new name of the field, "Accounted" (previously it was "InstanceOrArchetype")
        # (internally the values have the same meaning, "Instance" for a processor which has to be accounted,
        # "Archetype" for a processor which hasn't)
        v = attributes.get("instance_or_archetype", None)
        if strcmp(v, "Yes"):
            v = "Instance"
        elif strcmp(v, "No"):
            v = "Archetype"
        if v:
            attributes["instance_or_archetype"] = v

        name = field_values["processor"]
        p_names, _ = obtain_name_parts(name)

        geolocation = Geolocation.create(field_values["geolocation_ref"], field_values["geolocation_code"])

        ps = find_processors_matching_name(name, self._glb_idx)
        more_than_one = len(ps) > 1
        simple = len(p_names) == 1
        exists = True if len(ps) == 1 else False
        # SIMPLE? EXISTS? PARENT? ACTION:
        # Yes     Yes     Yes     NEW; HANG FROM PARENT
        # Yes     Yes     No      Warning: repeated
        # Yes     No      Yes     NEW; HANG FROM PARENT
        # Yes     No      No      NEW
        # No      Yes     Yes     Warning: cannot hang from parent
        # No      Yes     No      Warning: repeated AND not simple not allowed
        # No      No      Yes     Warning: cannot create more than one processor AND not simple not allowed
        # No      No      No      Warning: cannot create more than one processor AND not simple not allowed

        create_new = False
        if not simple:
            if not parent_processor:
                self._add_issue(IType.WARNING,
                                f"When a processor does not have parent, the name must be simple. Skipped." + subrow_issue_message(subrow))
                return
        else:
            if exists and not parent_processor:
                self._add_issue(IType.WARNING,
                                f"Repeated declaration of {name}. Skipped." + subrow_issue_message(subrow))
                return
            create_new = True

        if create_new:
            p = find_or_create_processor(
                state=self._glb_idx,
                name=name,
                proc_attributes=attributes,
                proc_location=geolocation)
        else:
            if exists:
                p = ps[0]

        # Add to ProcessorsGroup, if specified
        if pgroup:
            p_set = self._p_sets.get(pgroup, ProcessorsSet(pgroup))
            self._p_sets[pgroup] = p_set
            if p_set.append(p, self._glb_idx):  # Appends codes to the pset if the processor was not member of the pset
                p_set.append_attributes_codes(field_values["attributes"])

        # If geolocation specified, check if it exists
        # Inside it, check it the code exists
        if p.geolocation and p.geolocation.reference:
            # Geographical reference
            gr = self._glb_idx.get(GeographicReference.partial_key(name=p.geolocation.reference))
            if len(gr) == 0:
                self._add_issue(IType.ERROR, f"Geographical reference {p.geolocation.reference} not found "+subrow_issue_message(subrow))
                return
            if p.geolocation.reference and not p.geolocation.code:
                self._add_issue(IType.ERROR, f"Geographical reference was specified but not the code in it "+subrow_issue_message(subrow))
                return
            geo_id = p.geolocation.code
            try:
                url = gr[0].attributes["data_location"]
            except:
                self._add_issue(IType.ERROR, f"URL not found in geographical reference {p.geolocation.reference} "+subrow_issue_message(subrow))
                return
            try:
                j, ids = read_geojson(url)  # READ the file!! (or get it from cache). Could take some time...
            except:
                self._add_issue(IType.ERROR, f"URL {url} in reference {p.geolocation.reference} could not be read "+subrow_issue_message(subrow))
                return
            if geo_id not in ids:
                self._add_issue(IType.WARNING, f"Could not find code {geo_id} in file {url}, geographical reference {p.geolocation.reference} "+subrow_issue_message(subrow))

        # Add Relationship "part-of" if parent was specified
        # The processor may have previously other parent processors that will keep its parentship
        if parent_processor:
            # Create "part-of" relationship
            if len(self._glb_idx.get(ProcessorsRelationPartOfObservation.partial_key(parent_processor, p))) > 0:
                self._add_issue(IType.WARNING,
                                f"{p.name} is already part-of {parent_processor.name}. Skipped." + subrow_issue_message(subrow))
                return

            o1 = ProcessorsRelationPartOfObservation.create_and_append(parent_processor, p, None, behave_as=behave_as_processor, weight=field_values.get("parent_processor_weight"))  # Part-of
            self._glb_idx.put(o1.key(), o1)
            for hname in parent_processor.full_hierarchy_names(self._glb_idx):
                p_key = Processor.partial_key(f"{hname}.{p.name}", p.ident)
                if attributes:
                    p_key.update({k: ("" if v is None else v) for k, v in attributes.items()})
                self._glb_idx.put(p_key, p)




"""
Functionality to process an R script

* Prepare context for the execution:
  - special variables
  - state
* Execution:
  - the script, which has been designed to work accessing the RESTful API, will be executed locally
  - Instead of being an agent generating commands, it is itself a commands generator
  - Sandbox: no graphical functions, no filesystem operations, no output, only command execution API calls allowed (no NEW case study or similar, because the script will be in a case study), ...
  - no need to REGISTER, just MODIFY STATE. There must be a way for the endpoint to support this two behaviors

* After execution


"""


import json
from typing import Optional, Dict, Any

from nexinfosys.command_generators import Issue, IssueLocation, IType
from nexinfosys.common.helper import strcmp, first
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import Observer, FactorTypesRelationUnidirectionalLinearTransformObservation, \
    FactorType, Processor
from nexinfosys.command_executors import BasicCommand, CommandExecutionError, subrow_issue_message
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.models.musiasem_concepts_helper import find_or_create_observer, find_processor_by_name


class ScaleConversionV2Command(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))

    def _process_row(self, fields: Dict[str, Any], subrow=None) -> None:
        origin_interface_types = self._get_factor_types_from_field("source_hierarchy", "source_interface_type")
        destination_interface_types = self._get_factor_types_from_field("target_hierarchy", "target_interface_type")

        origin_processor: Optional[Processor] = None
        if fields["source_context"]:
            origin_processor = self._get_processor_from_field("source_context")

        destination_processor: Optional[Processor] = None
        if fields["target_context"]:
            destination_processor = self._get_processor_from_field("target_context")

        # Check that the interface types are from different hierarchies (warn if not; not error)
        for origin_interface_type in origin_interface_types:
            for destination_interface_type in destination_interface_types:
                if origin_interface_type.hierarchy == destination_interface_type.hierarchy:
                    self._add_issue(IType.WARNING, f"The interface types '{origin_interface_type.name}' and "
                                                   f"'{destination_interface_type.name}' are in the same hierarchy"+subrow_issue_message(subrow))

                observer = find_or_create_observer(Observer.no_observer_specified, self._glb_idx)

                # Check existence of ScaleChange
                key = FactorTypesRelationUnidirectionalLinearTransformObservation.partial_key(origin_interface_type, destination_interface_type, origin_processor, destination_processor, observer)
                tmp = self._glb_idx.get(key)
                if len(tmp) == 0:
                    # Create the directed Scale (Linear "Transformation") Relationship
                    o = FactorTypesRelationUnidirectionalLinearTransformObservation.create_and_append(
                        origin_interface_type, destination_interface_type, fields["scale"],
                        origin_processor, destination_processor,
                        fields["source_unit"], fields["target_unit"],
                        observer)
                    self._glb_idx.put(o.key(), o)
                elif len(tmp) > 0:
                    self._add_issue(IType.WARNING, f"The ScaleChange {tmp[0]} has already been defined with weight {tmp[0]._weight}")



import json

from nexinfosys import case_sensitive
from nexinfosys.common.helper import create_dictionary
from nexinfosys.model_services import IExecutableCommand, State, get_case_study_registry_objects
from nexinfosys.command_generators import parser_field_parsers
from nexinfosys.models.musiasem_concepts_helper import create_or_append_quantitative_observation
from nexinfosys.models.musiasem_concepts import FlowFundRoegenType, ProcessorsSet, PedigreeMatrix, Reference


class DataInputCommand(IExecutableCommand):
    """
    Serves to specify quantities (and their qualities) for observables
    If observables (Processor, Factor, FactorInProcessor) do not exist, they are created. This makes this command very powerfu: it may express by itself a MuSIASEM 1.0 structure

    """
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        """ The execution creates one or more Processors, Factors, FactorInProcessor and Observations
            It also creates "flat" Categories (Code Lists)
            It also Expands referenced Datasets
            Inserting it into "State"
        """
        def process_row(row):
            """
            Process a dictionary representing a row of the data input command. The dictionary can come directly from
            the worksheet or from a dataset.

            Implicitly uses "glb_idx"

            :param row: dictionary
            """
            # From "ff_type" extract: flow/fund, external/internal, incoming/outgoing
            # ecosystem/society?
            ft = row["ff_type"].lower()
            if ft == "int_in_flow":
                roegen_type = FlowFundRoegenType.flow
                internal = True
                incoming = True
            elif ft == "int_in_fund":
                roegen_type = FlowFundRoegenType.fund
                internal = True
                incoming = True
            elif ft == "ext_in_fund":
                roegen_type = FlowFundRoegenType.fund
                internal = False
                incoming = True
            elif ft == "int_out_flow":
                roegen_type = FlowFundRoegenType.flow
                internal = True
                incoming = False
            elif ft == "ext_in_flow":
                roegen_type = FlowFundRoegenType.flow
                internal = False
                incoming = True
            elif ft == "ext_out_flow":
                roegen_type = FlowFundRoegenType.flow
                internal = False
                incoming = False
            elif ft == "env_out_flow":
                roegen_type = FlowFundRoegenType.flow
                internal = False
                incoming = False
            elif ft == "env_in_flow":
                roegen_type = FlowFundRoegenType.flow
                internal = False
                incoming = True
            elif ft == "env_in_fund":
                roegen_type = FlowFundRoegenType.fund
                internal = False
                incoming = True

            # Split "taxa" attributes. "scale" corresponds to the observation
            p_attributes = row["taxa"].copy()
            if "scale" in p_attributes:
                other_attrs = create_dictionary()
                other_attrs["scale"] = p_attributes["scale"]
                del p_attributes["scale"]
            else:
                other_attrs = None

            # Check existence of PedigreeMatrix, if used
            if "pedigree_matrix" in row:
                pm = glb_idx.get(PedigreeMatrix.partial_key(name=row["pedigree_matrix"]))
                if len(pm) != 1:
                    issues.append((3, "Could not find Pedigree Matrix '"+row["pedigree_matrix"]+"'"))
                    del row["pedigree_matrix"]
                else:
                    try:
                        lst = pm[0].get_modes_for_code(row["pedigree"])
                    except:
                        issues.append((3, "Could not decode Pedigree '"+row["pedigree"]+"' for Pedigree Matrix '"+row["pedigree_matrix"]+"'"))
                        del row["pedigree"]
                        del row["pedigree_matrix"]
            else:
                if "pedigree" in row:
                    issues.append((3, "Pedigree specified without accompanying Pedigree Matrix"))
                    del row["pedigree"]

            # Source
            if "source" in row:
                try:
                    ast = parser_field_parsers.string_to_ast(parser_field_parsers.reference, row["source"])
                    ref_id = ast["ref_id"]
                    references = glb_idx.get(Reference.partial_key(ref_id), ref_type="provenance")
                    if len(references) == 1:
                        source = references[0]
                except:
                    source = row["source"]
            else:
                source = None

            # Geolocation
            if "geolocation" in row:
                try:
                    ast = parser_field_parsers.string_to_ast(parser_field_parsers.reference, row["geolocation"])
                    ref_id = ast["ref_id"]
                    references = glb_idx.get(Reference.partial_key(ref_id), ref_type="geographic")
                    if len(references) == 1:
                        geolocation = references[0]
                except:
                    geolocation = row["geolocation"]
            else:
                geolocation = None

            # CREATE FactorType, A Type of Observable, IF it does not exist
            # AND ADD Quantitative Observation
            p, ft, f, o = create_or_append_quantitative_observation(
                glb_idx,
                factor=row["processor"]+":"+row["factor"],
                value=row["value"] if "value" in row else None,
                unit=row["unit"],
                observer=source,
                spread=row["uncertainty"] if "uncertainty" in row else None,
                assessment=row["assessment"] if "assessment" in row else None,
                pedigree=row["pedigree"] if "pedigree" in row else None,
                pedigree_template=row["pedigree_matrix"] if "pedigree_matrix" in row else None,
                relative_to=row["relative_to"] if "relative_to" in row else None,
                time=row["time"] if "time" in row else None,
                geolocation=None,
                comments=row["comments"] if "comments" in row else None,
                tags=None,
                other_attributes=other_attrs,
                proc_aliases=None,
                proc_external=False,  # TODO
                proc_attributes=p_attributes,
                proc_location=None,
                ftype_roegen_type=roegen_type,
                ftype_attributes=None,
                fact_external=not internal,
                fact_incoming=incoming,
                fact_location=geolocation
            )
            if p_set.append(p, glb_idx):  # Appends codes to the pset if the processor was not member of the pset
                p_set.append_attributes_codes(row["taxa"])

            # author = state.get("_identity")
            # if not author:
            #     author = "_anonymous"
            #
            # oer = glb_idx.get(Observer.partial_key(author))
            # if not oer:
            #     oer = Observer(author, "Current user" if author != "_anonymous" else "Default anonymous user")
            #     glb_idx.put(oer.key(), oer)
            # else:
            #     oer = oer[0]
        # -----------------------------------------------------------------------------------------------------

        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

        # TODO Check semantic validity, elaborate issues
        issues = []

        p_set_name = self._name.split(" ")[1] if self._name.lower().startswith("processor") else self._name
        if self._name not in p_sets:
            p_set = ProcessorsSet(p_set_name)
            p_sets[p_set_name] = p_set
        else:
            p_set = p_sets[p_set_name]

        # Store code lists (flat "hierarchies")
        for h in self._content["code_lists"]:
            # TODO If some hierarchies already exist, check that they grow (if new codes are added)
            if h not in hh:
                hh[h] = []
            hh[h].extend(self._content["code_lists"][h])

        dataset_column_rule = parser_field_parsers.dataset_with_column

        processor_attributes = self._content["processor_attributes"]

        # Read each of the rows
        for i, r in enumerate(self._content["factor_observations"]):
            # Create processor, hierarchies (taxa) and factors
            # Check if the processor exists. Two ways to characterize a processor: name or taxa
            """
            ABOUT PROCESSOR NAME
            The processor can have a name and/or a set of qualifications, defining its identity
            If not defined, the name can be assumed to be the qualifications, concatenated
            Is assigning a name for all processors a difficult task? 
            * In the specification moment, it can get in the middle
            * When operating it is not so important
            * If taxa identify uniquely the processor, name is optional, automatically obtained from taxa
            * The benefit is that it can help reducing hierarchic names
            * It may help in readability of the case study
            
            """
            # If a row contains a reference to a dataset, expand it
            if "_referenced_dataset" in r:
                if r["_referenced_dataset"] in datasets:
                    ds = datasets[r["_referenced_dataset"]]  # Obtain dataset
                else:
                    ds = None
                    issues.append((3, "Dataset '" + r["_referenced_dataset"] + "' is not declared. Row "+str(i+1)))
            else:
                ds = None
            if ds:
                # Obtain a dict to map columns to dataset columns
                fixed_dict = {}
                var_dict = {}
                var_taxa_dict = {}
                for k in r:  # Iterate through columns in row "r"
                    if k == "taxa":
                        for t in r[k]:
                            if r[k][t].startswith("#"):
                                var_taxa_dict[t] = r[k][t][1:]
                        fixed_dict["taxa"] = r["taxa"].copy()
                    elif k in ["_referenced_dataset", "_processor_type"]:
                        continue
                    elif not r[k].startswith("#"):
                        fixed_dict[k] = r[k]  # Does not refer to the dataset
                    else:  # Starts with "#"
                        if k != "processor":
                            var_dict[k] = r[k][1:]  # Dimension
                        else:
                            fixed_dict[k] = r[k]  # Special
                # Check that the # names are in the Dataset
                if not case_sensitive:
                    s1 = {v.lower(): v for v in list(var_dict.values())+list(var_taxa_dict.values())}
                    s2 = {v.lower(): v for v in ds.data.columns}
                else:
                    s1 = {v: v for v in list(var_dict.values())+list(var_taxa_dict.values())}
                    s2 = {v: v for v in ds.data.columns}
                diff = set(s1.keys()).difference(set(s2.keys()))
                if diff:
                    # There are request fields in var_dict NOT in the input dataset "ds.data"
                    if len(diff) > 1:
                        v = "is"
                    else:
                        v = "are"
                    issues.append((3, "'"+', '.join(diff)+"' "+v+" not present in the requested dataset '"+r["_referenced_dataset"]+"'. Columns are: "+', '.join(ds.data.columns)+". Row " + str(i+1)))
                else:
                    # Iterate the dataset (a pd.DataFrame), row by row
                    for r_num, r2 in ds.data.iterrows():
                        r_exp = fixed_dict.copy()
                        if not case_sensitive:
                            r_exp.update({k: str(r2[s2[v.lower()]]) for k, v in var_dict.items()})
                        else:
                            r_exp.update({k: str(r2[s2[v]]) for k, v in var_dict.items()})
                        if var_taxa_dict:
                            taxa = r_exp["taxa"]
                            if not case_sensitive:
                                taxa.update({k: r2[s2[v.lower()]] for k, v in var_taxa_dict.items()})
                            else:
                                taxa.update({k: r2[s2[v]] for k, v in var_taxa_dict.items()})
                            if r_exp["processor"].startswith("#"):
                                r_exp["processor"] = "_".join([str(taxa[t]) for t in processor_attributes if t in taxa])
                        process_row(r_exp)
            else:  # Literal values
                process_row(r)

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the content
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues
